// TODO: Test basic mode
// #[test]
// fn test_basic_mode() {
//     let ret_bottom_layout = BottomLayout::init_basic_default(false);
// }

// TODO: Test moving around with procs and their hidden children.

// TODO: Test moving around with cpus if they get hidden.
